@@ -23,14 +23,20 @@ import pyqtgraph as pg
 from pyqtgraph import QtCore, QtGui
 
 
+
+# appropriate types
+ColorMap = pg.ColorMap
+
+# module definitions, mostly colors
 legend_border_color = '#000000dd'
 legend_fill_color   = '#00000055'
 legend_text_color   = '#dddddd66'
 soft_colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
 hard_colors = ['#000000', '#772211', '#000066', '#555555', '#0022cc', '#ffcc00']
-cmap_clash = pg.ColorMap([0.0, 0.2, 0.6, 1.0], [[127, 127, 255, 51], [0, 0, 127, 51], [255, 51, 102, 51], [255, 178, 76, 51]])
+colmap_clash = ColorMap([0.0, 0.2, 0.6, 1.0], [[127, 127, 255, 51], [0, 0, 127, 51], [255, 51, 102, 51], [255, 178, 76, 51]])
 foreground = '#000000'
 background = '#ffffff'
+odd_plot_background = '#f0f0f0'
 candle_bull_color = '#26a69a'
 candle_bear_color = '#ef5350'
 candle_bull_body_color = background
@@ -39,7 +45,6 @@ volume_bear_color = '#f7a9a7'
 volume_bull_body_color = volume_bull_color
 volume_neutral_color = '#b0b0b0'
 poc_color = '#000060'
-odd_plot_background = '#f0f0f0'
 band_color = '#d2dfe6'
 cross_hair_color = '#00000077'
 draw_line_color = '#000000'
@@ -126,23 +131,23 @@ class YAxisItem(pg.AxisItem):
 class YScale:
     def __init__(self, scaletype, scalef):
         self.scaletype = scaletype
-        self.scalef = scalef
+        self.set_scale(scalef)
 
     def set_scale(self, scale):
-        self.scalef = scale if self.scaletype != 'log' else 1
+        self.scalef = scale
 
     def xform(self, y):
         if self.scaletype == 'log':
             y = 10**y
-        return y * self.scalef
+        y = y * self.scalef
+        return y
 
     def invxform(self, y, verify=False):
+        y /= self.scalef
         if self.scaletype == 'log':
             if verify and y <= 0:
                 return -1e6
             y = np.log10(y)
-        else:
-            y /= self.scalef
         return y
 
 
@@ -283,13 +288,13 @@ class PandasDataSource:
         orig_cols = list(df.columns)
         timecol,orig_cols = orig_cols[0],orig_cols[1:]
         df = df.set_index(timecol)
-        data = datasrc.df.set_index(datasrc.df.columns[0])
-        data.columns = [self.renames.get(col, col) for col in data.columns]
+        input_df = datasrc.df.set_index(datasrc.df.columns[0])
+        input_df.columns = [self.renames.get(col, col) for col in input_df.columns]
         for col in df.columns:
-            if col not in data.columns:
-                data[col] = df[col]
-        data = data.reset_index()
-        self.df = data[[data.columns[0]]+orig_cols] if orig_cols else data
+            if col not in input_df.columns:
+                input_df[col] = df[col]
+        input_df = input_df.reset_index()
+        self.df = input_df[[input_df.columns[0]]+orig_cols] if orig_cols else input_df
         self.init_x1 = self.xlen - side_margin
         self.cache_hilo = OrderedDict()
         self._period = None
@@ -583,6 +588,7 @@ class FinViewBox(pg.ViewBox):
     def __init__(self, win, init_steps=300, yscale=YScale('linear', 1), v_zoom_scale=1, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.win = win
+        self.master_viewbox = None
         self.yscale = yscale
         self.v_zoom_scale = v_zoom_scale
         self.v_zoom_baseline = 0.5
@@ -607,6 +613,14 @@ class FinViewBox(pg.ViewBox):
             return
         datasrc.update_init_x(self.init_steps)
 
+    def pre_process_data(self):
+        if self.datasrc and self.datasrc.scale_cols:
+            df = self.datasrc.df.iloc[:, self.datasrc.scale_cols]
+            self.y_max = df.max().max()
+            self.y_min = df.min().min()
+            if self.y_min <= 0:
+                self.y_positive = False
+
     @property
     def datasrc_or_standalone(self):
         ds = self.datasrc
@@ -615,6 +629,8 @@ class FinViewBox(pg.ViewBox):
         return ds
 
     def wheelEvent(self, ev, axis=None):
+        if self.master_viewbox:
+            return self.master_viewbox.wheelEvent(ev, axis=axis)
         if ev.modifiers() == QtCore.Qt.ControlModifier:
             scale_fact = 1
             self.v_zoom_scale /= 1.02 ** (ev.delta() * self.state['wheelScaleFactor'])
@@ -632,6 +648,8 @@ class FinViewBox(pg.ViewBox):
         ev.accept()
 
     def mouseDragEvent(self, ev, axis=None):
+        if self.master_viewbox:
+            return self.master_viewbox.mouseDragEvent(ev, axis=axis)
         if not self.datasrc:
             return
         if ev.button() == QtCore.Qt.LeftButton:
@@ -696,6 +714,8 @@ class FinViewBox(pg.ViewBox):
         ev.accept()
 
     def mouseClickEvent(self, ev):
+        if self.master_viewbox:
+            return self.master_viewbox.mouseClickEvent(ev)
         if _mouse_clicked(self, ev):
             ev.accept()
             return
@@ -709,6 +729,8 @@ class FinViewBox(pg.ViewBox):
         ev.accept()
 
     def keyPressEvent(self, ev):
+        if self.master_viewbox:
+            return self.master_viewbox.keyPressEvent(ev)
         if _key_pressed(self, ev):
             ev.accept()
             return
@@ -958,10 +980,10 @@ class CandlestickItem(FinPlotItem):
 
 
 class HeatmapItem(FinPlotItem):
-    def __init__(self, ax, datasrc, rect_size=0.9, filter_limit=0, cmap=cmap_clash, whiteout=0.0, colcurve=lambda x:pow(x,4)):
+    def __init__(self, ax, datasrc, rect_size=0.9, filter_limit=0, colmap=colmap_clash, whiteout=0.0, colcurve=lambda x:pow(x,4)):
         self.rect_size = rect_size
         self.filter_limit = filter_limit
-        self.cmap = cmap
+        self.colmap = colmap
         self.whiteout = whiteout
         self.colcurve = colcurve
         self.col_data_end = len(datasrc.df.columns)
@@ -985,7 +1007,7 @@ class HeatmapItem(FinPlotItem):
                 v = row[ci]
                 if v >= lim:
                     v = 1 - self.colcurve(1 - (v-lim)/(1-lim))
-                    color = self.cmap.map(v, mode='qcolor')
+                    color = self.colmap.map(v, mode='qcolor')
                     p.fillRect(QtCore.QRectF(t-rect_size2, price+h0, self.rect_size, h1), color)
         p.end()
 
@@ -1243,12 +1265,8 @@ def renko(x, y=None, bins=None, step=None, ax=None, colorfunc=price_colorfilter)
     ax = _create_plot(ax=ax, maximize=False)
     datasrc = _create_datasrc(ax, x, y)
     origdf = datasrc.df
-    if not bins and not step:
-        bins = 50
-    if bins:
-        step = (datasrc.y.max()-datasrc.y.min()) / bins
     adj = _adjust_renko_log_datasrc if ax.vb.yscale.scaletype == 'log' else _adjust_renko_datasrc
-    step_adjust_renko_datasrc = partial(adj, step)
+    step_adjust_renko_datasrc = partial(adj, bins, step)
     step_adjust_renko_datasrc(datasrc)
     ax.decouple()
     item = candlestick_ochl(datasrc, draw_shadow=False, candle_width=1, ax=ax, colorfunc=colorfunc)
@@ -1348,7 +1366,7 @@ def bar(x, y=None, width=0.8, ax=None, colorfunc=strength_colorfilter):
     _adjust_bar_datasrc(datasrc, order_cols=False) # don't rearrange columns, done for us in volume_ocv()
     item = volume_ocv(datasrc, candle_width=width, ax=ax, colorfunc=colorfunc)
     item.update_data = partial(_update_data, None, _adjust_bar_datasrc, item)
-    _pre_process_data(ax.vb)
+    ax.vb.pre_process_data()
     if ax.vb.y_min >= 0:
         ax.vb.v_zoom_baseline = 0
     return item
@@ -1546,7 +1564,7 @@ def show(qt_exec=True):
     for win in windows:
         vbs = [ax.vb for ax in win.axs]
         for vb in vbs:
-            _pre_process_data(vb)
+            vb.pre_process_data()
         if viewrestore:
             if _loadwindata(win):
                 continue
@@ -1682,14 +1700,22 @@ def _add_timestamp_plot(master, prev_ax, viewbox, index, yscale):
     ax.decouple = partial(_ax_decouple, ax)
     ax.disable_x_index = partial(_ax_disable_x_index, ax)
     ax.prev_ax = prev_ax
+    ax.win_index = index
     if index%2:
         viewbox.setBackgroundColor(odd_plot_background)
     viewbox.setParent(ax)
     return ax
 
 
-def _overlay(ax, scale=0.25):
-    viewbox = FinViewBox(ax.vb.win, init_steps=ax.vb.init_steps, yscale=YScale('linear', 1))
+def _overlay(ax, scale=0.25, y_axis=False):
+    '''The scale parameter defines how "high up" on the initial plot this overlay will show.
+       The y_axis parameter can be one of [False, 'linear', 'log'].'''
+    yscale = y_axis if y_axis else 'linear'
+    viewbox = FinViewBox(ax.vb.win, init_steps=ax.vb.init_steps, yscale=YScale(yscale, 1))
+    viewbox.master_viewbox = ax.vb
+    viewbox.setZValue(-5)
+    viewbox.setBackgroundColor(ax.vb.state['background'])
+    ax.vb.setBackgroundColor(None)
     viewbox.v_zoom_scale = scale
     if hasattr(ax, 'ax_widget'):
         ax.ax_widget.scene().addItem(viewbox)
@@ -1709,6 +1735,15 @@ def _overlay(ax, scale=0.25):
     axo.hideAxis('bottom')
     axo.hideButtons()
     viewbox.addItem(axo)
+    if y_axis and isinstance(axo.vb.win, pg.GraphicsLayoutWidget):
+        axi = YAxisItem(vb=axo.vb, orientation='right')
+        axo.axes['right'] = {'item':axi}
+        axi.linkToView(axo.vb)
+        row = ax.win_index
+        for col in range(1, 100):
+            if axo.vb.win.getItem(row, col) is None:
+                axo.vb.win.addItem(axi, row=row, col=1)
+                break
     ax.vb.sigResized.connect(updateView)
     overlay_axs.append(axo)
     updateView()
@@ -1805,10 +1840,16 @@ def _set_datasrc(ax, datasrc):
             viewbox.set_datasrc(datasrc) # for mwheel zoom-scaling
             _set_x_limits(ax, datasrc)
         else:
+            t0 = viewbox.datasrc.x.loc[0]
             viewbox.datasrc.addcols(datasrc)
+            # check if we need to re-render previous plots due to changed indices
+            indices_updated = viewbox.datasrc.timebased() and t0 != viewbox.datasrc.x.loc[0]
             for item in ax.items:
-                if isinstance(item, FinPlotItem):
+                if hasattr(item, 'datasrc'):
                     item.datasrc.set_df(viewbox.datasrc.df) # every plot here now has the same time-frame
+                    if indices_updated:
+                        _start_visual_update(item)
+                        _end_visual_update(item)
             _set_x_limits(ax, datasrc)
             viewbox.set_datasrc(viewbox.datasrc) # update zoom
     else:
@@ -1816,9 +1857,9 @@ def _set_datasrc(ax, datasrc):
         datasrc.update_init_x(viewbox.init_steps)
         ## if not viewbox.x_indexed:
             ## _set_x_limits(ax, datasrc)
-    # update period if this datasrc has higher resolution
+    # update period if this datasrc has higher time resolution
     global epoch_period
-    if epoch_period > 1e7 or not datasrc.standalone:
+    if datasrc.timebased() and (epoch_period > 1e7 or not datasrc.standalone):
         ep = datasrc.period
         epoch_period = ep if ep < epoch_period else epoch_period
 
@@ -1827,7 +1868,11 @@ def _has_timecol(df):
     return len(df.columns) >= 2
 
 
-def _adjust_renko_datasrc(step, datasrc):
+def _adjust_renko_datasrc(bins, step, datasrc):
+    if not bins and not step:
+        bins = 50
+    if not step:
+        step = (datasrc.y.max()-datasrc.y.min()) / bins
     bricks = datasrc.y.diff() / step
     bricks = (datasrc.y[bricks.isnull() | (bricks.abs()>=0.5)] / step).round().astype(int)
     extras = datasrc.df.iloc[:, datasrc.col_data_offset+1:]
@@ -1855,11 +1900,9 @@ def _adjust_renko_datasrc(step, datasrc):
     datasrc.set_df(pd.DataFrame(data, columns='time open close high low'.split()+list(extras.columns)))
 
 
-def _adjust_renko_log_datasrc(step, datasrc):
-    bins = (datasrc.y.max()-datasrc.y.min()) / step
+def _adjust_renko_log_datasrc(bins, step, datasrc):
     datasrc.df.iloc[:,1] = np.log10(datasrc.df.iloc[:,1])
-    step = (datasrc.y.max()-datasrc.y.min()) / bins
-    _adjust_renko_datasrc(step, datasrc)
+    _adjust_renko_datasrc(bins, step, datasrc)
     datasrc.df.iloc[:,1:5] = 10**datasrc.df.iloc[:,1:5]
 
 
@@ -1911,13 +1954,8 @@ def _update_data(preadjustfunc, adjustfunc, item, ds):
     ds = _create_datasrc(item.ax, ds)
     if adjustfunc:
         adjustfunc(ds)
-    if isinstance(item, FinPlotItem):
-        item.ax.removeItem(item)
-        item.dirty = True
-        item.datasrc.update(ds)
-    else:
-        item.datasrc.update(ds)
-        item.setData(item.datasrc.index, item.datasrc.y)
+    item.datasrc.update(ds)
+    _start_visual_update(item)
     for i in item.ax.items:
         if i == item or not hasattr(i, 'datasrc'):
             continue
@@ -1952,22 +1990,23 @@ def _update_data(preadjustfunc, adjustfunc, item, ds):
         if item.ax.axes['bottom']['item'].isVisible(): # update axes if visible
             item.ax.axes['bottom']['item'].hide()
             item.ax.axes['bottom']['item'].show()
-    for ax in item.ax.vb.win.axs:
-        ax.vb.update()
-    if isinstance(item, FinPlotItem):
-        item.ax.addItem(item)
-        item.repaint()
+    _end_visual_update(item)
     if update_sigdig:
         _update_significants(item.ax, item.ax.vb.datasrc, force=True)
 
 
-def _pre_process_data(vb):
-    if vb.datasrc and vb.datasrc.scale_cols:
-        df = vb.datasrc.df.iloc[:, vb.datasrc.scale_cols]
-        vb.y_max = df.max().max()
-        vb.y_min = df.min().min()
-        if vb.y_min <= 0:
-            vb.y_positive = False
+def _start_visual_update(item):
+    if isinstance(item, FinPlotItem):
+        item.ax.removeItem(item)
+        item.dirty = True
+    else:
+        item.setData(item.datasrc.index, item.datasrc.y)
+
+
+def _end_visual_update(item):
+    if isinstance(item, FinPlotItem):
+        item.ax.addItem(item)
+        item.repaint()
 
 
 def _set_x_limits(ax, datasrc):
@@ -2087,6 +2126,8 @@ def _inspect_clicked(ax, inspector, evs):
 
 
 def _inspect_pos(ax, inspector, poss):
+    if not ax.vb.datasrc:
+        return
     point = ax.vb.mapSceneToView(poss[-1])
     t = point.x() + 0.5
     try:
@@ -2107,7 +2148,7 @@ def brighten(color, f):
 
 
 def _get_color(ax, style, wanted_color):
-    if type(wanted_color) == str:
+    if type(wanted_color) in (str, QtGui.QColor):
         return wanted_color
     index = wanted_color if type(wanted_color) == int else None
     if style is None or any(ch in style for ch in '-_.'):
@@ -2127,9 +2168,10 @@ def _pdtime2epoch(t):
         if isinstance(t.iloc[0], pd.Timestamp):
             return t.astype('int64') // int(1e6)
         if np.nanmax(t.values) > 1e13: # handle ns epochs
-            return t.astype('float64') / 1e3
+            return (t/1e3).astype('int64')
         if np.nanmax(t.values) < 1e10: # handle s epochs
-            return t.astype('float64') * 1e3
+            return (t*1e3).astype('int64')
+        return t.astype('int64')
     return t
 
 
@@ -2188,7 +2230,7 @@ def _get_datasrc(ax, require=True):
 
 
 def _x2local_t(datasrc, x):
-    return _x2t(datasrc, x, lambda t: datetime.fromtimestamp(t/1000, tz=display_timezone).isoformat(sep=' ').split('+')[0])
+    return _x2t(datasrc, x, lambda t: datetime.fromtimestamp(t/1000, tz=display_timezone).isoformat(sep=' ').partition('+')[0])
 
 
 def _x2utc(datasrc, x):
@@ -2205,9 +2247,9 @@ def _x2t(datasrc, x, ts2str):
             if not datasrc.timebased():
                 return '%g' % t, False
             s = ts2str(t)
-            if epoch_period >= 24*60*60:
+            if epoch_period >= 23*60*60: # daylight savings, leap seconds, etc
                 i = s.index(' ')
-            elif epoch_period >= 60:
+            elif epoch_period >= 59: # consider leap seconds
                 i = s.rindex(':')
             elif epoch_period >= 1:
                 i = s.index('.') if '.' in s else len(s)
@@ -2228,7 +2270,7 @@ def _x2year(datasrc, x):
 
 
 def _round_to_significant(rng, rngmax, x, significant_decimals, significant_eps):
-    is_highres = rng/significant_eps > 1e2 and (rngmax>1e7 or rngmax<1e-2)
+    is_highres = (rng/significant_eps > 1e2 and rngmax<1e-2) or abs(rngmax) > 1e7
     sd = significant_decimals
     if is_highres and abs(x)>0:
         exp10 = floor(np.log10(abs(x)))
@@ -2276,8 +2318,20 @@ def _clamp_point(ax, p):
 
 
 def _draw_line_segment_text(polyline, segment, pos0, pos1):
+    fsecs = None
+    datasrc = polyline.vb.datasrc
+    if datasrc and clamp_grid:
+        try:
+            x0,x1 = pos0.x()+0.5, pos1.x()+0.5
+            t0,_,_,_,cnt0 = datasrc.hilo(x0, x0)
+            t1,_,_,_,cnt1 = datasrc.hilo(x1, x1)
+            if cnt0 and cnt1:
+                fsecs = abs(t1 - t0) / 1000
+        except:
+            pass
     diff = pos1 - pos0
-    fsecs = abs(diff.x()*epoch_period)
+    if fsecs is None:
+        fsecs = abs(diff.x()*epoch_period)
     secs = int(fsecs)
     mins = secs//60
     hours = mins//60
